/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sun Mar/04/2001 03:23 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#include "Kanal.h"
#include "fw_cost.h"

using namespace std;
using namespace rs;
using namespace fw_cost;

bool Kanal::correct_date = false;

//  ----------------------------------------------------------------------------------
//      bool Connection::isInZeitraum(const time_t& von, const time_t& bis) const
//  ----------------------------------------------------------------------------------
bool Connection::isInZeitraum(const time_t& von, const time_t& bis) const {
    struct tm tm2             = *getStart();
    time_t    t1              = mktime(&tm2);
    int       t_dur           = getDuration();
    int       vb_dur          = int(difftime(bis, von));
    int       start_after_von = int(difftime(t1, von));
    int       start_after_bis = start_after_von-vb_dur;
    int       end_after_von   = start_after_von+t_dur;
    int       end_after_bis   = start_after_bis+t_dur;

    if (end_after_von<=0 || start_after_bis>=0) return false; // completely outside von-bis

    int outside = 0;
    if (start_after_von<0) { // starts before von
        outside += -start_after_von;
    }
    if (end_after_bis>0) { // ends after bis
        outside += end_after_bis;
    }

    assert(t_dur>=outside);
    int inside = t_dur-outside;

    return inside>=outside;
}

//  --------------------------------------------------------------------------------------------
//      void Kanal::parseDateTime(struct tm *t, const string& datum, const string& uhrzeit)
//  --------------------------------------------------------------------------------------------
void Kanal::parseDateTime(struct tm *t, const string& datum, const string& uhrzeit) {
    t->tm_year  = atoi(datum.substr(0, 4).c_str())-1900;
    t->tm_mon   = atoi(datum.substr(4, 2).c_str())-1;
    t->tm_mday  = atoi(datum.substr(6, 2).c_str());

    t->tm_hour  = atoi(uhrzeit.substr(0, 2).c_str());
    t->tm_min   = atoi(uhrzeit.substr(2, 2).c_str());
    t->tm_sec   = atoi(uhrzeit.substr(4, 2).c_str());

    t->tm_wday  = -1;
    t->tm_yday  = -1;
    t->tm_isdst = -1;

#ifdef MKTIME_TEST
    printf("-------------------- before mktime [1]:\n");
    printf("datum='%s'\nuhrzeit='%s'\n", datum.c_str(), uhrzeit.c_str());
    printf("t->tm_sec=%i\n", t->tm_sec);
    printf("t->tm_min=%i\n", t->tm_min);
    printf("t->tm_hour=%i\n", t->tm_hour);
    printf("t->tm_mday=%i\n", t->tm_mday);
    printf("t->tm_mon=%i\n", t->tm_mon);
    printf("t->tm_year=%i\n", t->tm_year);
    printf("t->tm_wday=%i\n", t->tm_wday);
    printf("t->tm_yday=%i\n", t->tm_yday);
    printf("t->tm_isdst=%i\n", t->tm_isdst);
#endif //MKTIME_TEST

    time_t ti = mktime(t);

#ifdef MKTIME_TEST
    printf("-------------------- after mktime [1]:\n");
    printf("ti=%lu\n", (long)ti);
    printf("t->tm_sec=%i\n", t->tm_sec);
    printf("t->tm_min=%i\n", t->tm_min);
    printf("t->tm_hour=%i\n", t->tm_hour);
    printf("t->tm_mday=%i\n", t->tm_mday);
    printf("t->tm_mon=%i\n", t->tm_mon);
    printf("t->tm_year=%i\n", t->tm_year);
    printf("t->tm_wday=%i\n", t->tm_wday);
    printf("t->tm_yday=%i\n", t->tm_yday);
    printf("t->tm_isdst=%i\n", t->tm_isdst);
#endif //MKTIME_TEST

    if (ti!=-1 && correct_date) {
        t->tm_mday++;
        ti = mktime(t);
    }

    if (ti==-1) throw rs::err::Error(string("Illegal date oder time ('")+datum+"', '"+uhrzeit+"')");
}

namespace fw_cost {

    //  ---------------------------------------------------------------------------------------------------------------
    //      ConnIter findConnectionSeries(ConnIter begin, ConnIter end, int max_pause, bool allowConnectionChange, bool& usesBothChannels)
    //  ---------------------------------------------------------------------------------------------------------------
    // searches from begin to end, until a pause longer than max_pause is detected
    // returns an iterator to the last Connection before that pause (end if no such pause)
    //
    ConnIter findConnectionSeries(ConnIter begin, ConnIter end, int max_pause, bool allowConnectionChange, bool& usesBothChannels) {
        ConnIter previous = begin;
        usesBothChannels  = false;

        for (++begin; begin!=end; ++begin) {
            if (begin->getKanal() == 0) {

                int pause  = int(*begin-*previous);
                pause     -= previous->getDuration();

                if (pause<0) break; // => Zeit verstellt
                if (pause>max_pause) break;

                if (!allowConnectionChange && begin->getVerbindung().getName()!=previous->getVerbindung().getName()) {
                    break; // Unterschiedliche Verbindungen
                }

                previous = begin;
            }
            else {
                usesBothChannels = true;
            }
        }

        // now 'begin' points to first connection behind series
        // set 'previous' to first before 'begin'

        ConnIter up = previous;
        for (++up; up!=begin; ++up) {
            previous = up;
        }

        return previous;
    }

};


