/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sat Dec/07/2002 13:53 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

// Inifile :
//
// C:\UT\WinBatches\fw_cost.ini

#include <cstdio>

#include <algorithm>
#include <functional>

#include <Option.h>
#include <dbf.h>

#include "fw_cost.h"
#include "read.h"
#include "output.h"
#include "store.h"

using namespace std;
using namespace fw_cost;
using namespace rs;
using namespace rs::ini;
using namespace rs::cmdline;
using namespace rs::err;
using namespace rs::dbf;
using namespace rs::date;
using namespace rs::str;

//  -----------------------
//      User-Interface:
//  -----------------------

static ProgramInfo fw_cost_info("fw_cost", 1, 9, "Kostenberechnung und Connection-Check fuer Fritz!web");

static CmdLineOption einzel('e', "Einzelne Verbindungen auflisten");
static CmdLineOption provider('P', "Verbindungen fuer Provider getrennt auflisten");

static CmdLineOption tagessummen('t', "Tagessummen auflisten");
static CmdLineOption wochensummen('w', "Wochensummen auflisten");
static CmdLineOption monatssummen('m', "Monatssummen auflisten");

static CmdLineOption stundenStatistik('U', "Statistik fr Uhrzeit");
static CmdLineOption wochentagStatistik('W', "Statistik fr Wochentage");

static CmdLineOption bundles('b', "Mehrere kurz aufeinanderfolgende Verbindungen werden gebuendelt "
                             "ausgegeben. Dabei wird berechnet, welche Einsparung sich gegenueber einer "
                             "ununterbrochenen Online-Verbindung ergibt.");

static CmdLineArgumentOption maxPause('p', "sec", "Maximale Laenge der Pausen innerhalb von "
                                      "\"kurz aufeinanderfolgenden\" Verbindungen.", "300");
static CmdLineOption allowConnectionChange('c', "Verbindungswechsel innerhalb gebuendelter Bloecke erlauben.");
static CmdLineOption detailed('d', "\"Kurz aufeinanderfolgende\" Verbindungen detailliert ausgeben.");

static CmdLineArgumentOption Zeitraum('z', "zeitraum", "Zeitraumsummen ausgeben. "
                                      "Syntax von <zeitraum> ist dd.mm.yyyy-dd.mm.yyyy", "all");

static CmdLineAnnotation empty0("", false);

static CmdLineArgumentOption testOnline('o', "ini", "Testet ob Fritzweb online ist und schreibt das Ergebnis in <ini>. ");

static CmdLineAnnotation empty1("", false);

static CmdLineOption         currentCosts('C', "Berechnet die momentanen Kosten fr die einzelnen Provider (Provider mit Faktor >99 werden ignoriert).");
static CmdLineArgumentOption forTime('T', "sec", (string("Zeitspanne fr -")+currentCosts.get_optChar()).c_str(), "59");

static CmdLineAnnotation empty2("", false);

static CmdLineArgumentOption queryLongConnections('q', "min",
                                                  "Manuelle Korrektur aller Verbindungen, "
                                                  "die lnger als <min> Minuten gedauert haben.");

static CmdLineAnnotation empty3("", false);

static CmdLineArgumentOption iniFile('i', "filename", "Alternatives Inifile", "fw_cost.ini");

static CmdLineAnnotation empty4("", false);

static CmdLineArgumentOption providerSucks('S', "add", "Addiere <add> zum Faktor des momentanen Providers.");

//  ---------------------
//      class Global
//  ---------------------
class Global {
public:
    Date today;
    //     struct tm today;

    Global() : today() {
//         time_t now;
//         time(&now);
//         today = Date(*localtime(&now));
    }

    virtual ~Global() {}
};

static Global global;

// --------------------------------------------------------------------------------
//     static void completeInifile(InifilePtr ini)
// --------------------------------------------------------------------------------
static void completeInifile(InifilePtr ini) {
    SectionPtr Main     = ini->findOrCreateSection("Allgemein");
    EntryPtr   event_db = Main->findEntry("EventDatenbank");
    EntryPtr   fritz_db =
        Main->findOrCreateEntry("FritzWebDatenbank",
                                event_db.Null()
                                ? "C:\\Programme\\FRITZ!\\FRIWEBEV.DBF"
                                : Main->findStringEntry("EventDatenbank"),
                                MAJOR_CHANGES,
                                "Mgliche Positionen der FritzWebDatenbank (AVM ist leicht unschluessig):\n"
                                "C:\\WINDOWS\\Anwendungsdaten\\FRITZ!\\FRIWEBEV.DBF\n"
                                "C:\\Programme\\FRITZ!\\FRIWEBEV.DBF\n"
                                "C:\\WINDOWS\\User\\...\\Anwendungsdaten\\FRITZ!\\FRIWEBEV.DBF"
                                );

    if (!event_db.Null()) event_db->mark_obsolete(); // "EventDatenbank" is obsolete

    bool convert_DM_to_Euro                = false;
    bool convert_EuroPerTakt_to_CentPerMin = false;

    {
        EntryPtr Euro = Main->findEntry("Euro");
        if (Euro.Null()) {
            Euro               = Main->findOrCreateEntry("Euro", 1, MAJOR_CHANGES,
                                                         "All prices in fw_cost.ini have been converted from DM to Euro");
            convert_DM_to_Euro = true;
        }
        EntryPtr CentPerMinute = Main->findEntry("CentPerMinute");
        if (CentPerMinute.Null()) {
            CentPerMinute                     = Main->findOrCreateEntry("CentPerMinute", 1, MAJOR_CHANGES,
                                                                        "All prices in fw_cost.ini have been converted from Euro/Takt to Cent/min");
            convert_EuroPerTakt_to_CentPerMin = true;
        }
    }

    if (convert_EuroPerTakt_to_CentPerMin || convert_DM_to_Euro) {
        for (Sections::const_iterator s = ini->begin(); s != ini->end(); ++s) {
            SectionPtr sec = s->second;

            for (Entries::const_iterator e = sec->begin(); e != sec->end(); ++e) {
                EntryPtr      entry = e->second;
                const string& name  = entry->getName();

                if (name.length()>8 && name.substr(name.length()-8) == "_Taktung") {
                    string content = sec->findStringEntry(name);
                    size_t komma   = content.find(',');

                    printf("%s=%s\n", name.c_str(), content.c_str());

                    if (komma != string::npos) {
                        string taktlaenge = content.substr(0, komma);
                        string preis      = content.substr(komma+1);

                        int    sekunden    = atoi(taktlaenge.c_str());
                        double dm          = strtod(preis.c_str(), 0);
                        double euroProTakt = convert_DM_to_Euro
                            ? (dm/1.95583) // convert DM -> Euro
                            : dm; // already is in Euro

                        double euroPerSecond = euroProTakt/double(sekunden);
                        double centPerMinute = convert_EuroPerTakt_to_CentPerMin
                            ? euroPerSecond*100.0*60.0
                            : euroProTakt; // price already was in cent/min

//                         string rest = content.substr(komma+1);
//                         double dm   = strtod(rest.c_str(), 0); // atof(rest.c_str());
//                         double euro = dm/1.95583;

                        content.erase(komma+1);
                        content += strf("%lf", centPerMinute);

                        // sec->updateOrCreateStringEntry(name, content);
                        entry->setContent(content);
                    }

                    printf("%s=%s\n", name.c_str(), content.c_str());
                }
            }
        }
    }

    ini->findOrCreateSection("Verbindungen");
}

//  -----------------------------------------------------------------
//      static bool isWordDelimit(char c, const string& delimit)
//  -----------------------------------------------------------------
static bool isWordDelimit(char c, const string& delimit) {
    return find(delimit.begin(), delimit.end(), c)!=delimit.end();
}
// --------------------------------------------------------------------------------
//     inline string getWord(string::iterator& s, string::iterator end, const string& delimit)
// --------------------------------------------------------------------------------
inline string getWord(string::iterator& s, string::iterator end, const string& delimit) {
    while (s!=end && isWordDelimit(*s, delimit)) ++s;
    string::iterator start = s;
    while (s!=end && !isWordDelimit(*s, delimit)) ++s;
    string::iterator after = s;
    --after;

    return string(start, after);
}


//  ------------------------------------------------------------------------
//      static bool sameDay(const Connection& c1, const Connection& c2)
//  ------------------------------------------------------------------------
static bool sameDay(const Connection& c1, const Connection& c2) {
    const struct tm * t1 = c1.getStart();
    const struct tm * t2 = c2.getStart();

    return
        t1->tm_mday == t2->tm_mday &&
        t1->tm_mon  == t2->tm_mon  &&
        t1->tm_year == t2->tm_year;
}
//  --------------------------------------------------------------------------
//      static bool sameMonth(const Connection& c1, const Connection& c2)
//  --------------------------------------------------------------------------
static bool sameMonth(const Connection& c1, const Connection& c2) {
    const struct tm * t1 = c1.getStart();
    const struct tm * t2 = c2.getStart();

    return
        t1->tm_mon  == t2->tm_mon &&
        t1->tm_year == t2->tm_year;
}

//  --------------------------------------------------
//      static int weekNumber(const struct tm *t)
//  --------------------------------------------------
static int weekNumber(const struct tm *t) {
    // Woche startet am Montag
    // 1. Woche des Jahres enthlt den 1. Donnerstag des Jahres

    int daysSinceSunday          = t->tm_wday; // 0=So 4=Do
    int dayOfTheYear             = t->tm_yday;
    int jan1_daysSinceSunday     = (daysSinceSunday-((dayOfTheYear-1)%7)+7)%7;
    int week1start_dayOfTheYear  = 1-jan1_daysSinceSunday+1;
    if (jan1_daysSinceSunday>4) { //
        week1start_dayOfTheYear += 7;
    }

    assert(week1start_dayOfTheYear>=-2 && week1start_dayOfTheYear<=4);
    // -2 falls 1.Jan==Do -> Woche1 startet am 29.12 des Vorjahres
    //  4 falls 1.Jan==Fr -> Woche1 startet am 04.01

    int daySinceStartOfWeek1 = dayOfTheYear-week1start_dayOfTheYear;
    int weeknr               = (daySinceStartOfWeek1/7) + 1;

    if (weeknr==0) weeknr  = 52;
    if (weeknr==53) weeknr = 1;
    return weeknr;

//     char buf[10];
//     strftime(buf, 9, "%W", t);
//     return atoi(buf);
}

//  -------------------------------------------------------------------------
//      static bool sameWeek(const Connection& c1, const Connection& c2)
//  -------------------------------------------------------------------------
static bool sameWeek(const Connection& c1, const Connection& c2) {
    return weekNumber(c1.getStart())==weekNumber(c2.getStart());
}

//  ----------------------------------------------------------------
//      static void writeDay(ostream& out, const Connection& c)
//  ----------------------------------------------------------------
static void writeDay(ostream& out, const Connection& c) {
    out << dateString(c.getStart()) << ": ";
}
//  ------------------------------------------------------------------
//      static void writeMonth(ostream& out, const Connection& c)
//  ------------------------------------------------------------------
static void writeMonth(ostream& out, const Connection& c) {
    out << dateString(c.getStart()).c_str()+3 << ": ";
}
//  -----------------------------------------------------------------
//      static void writeWeek(ostream& out, const Connection& c)
//  -----------------------------------------------------------------
static void writeWeek(ostream& out, const Connection& c) {
    out << "Woche " << rs::str::strf("%2i", weekNumber(c.getStart())) << ": ";
}

// --------------------------------------------------------------------------------
//     static void parseDateTime(struct tm *t, const string& datum, const string& uhrzeit)
// --------------------------------------------------------------------------------
static void parseDateTime(struct tm *t, const string& datum, const string& uhrzeit) {
    while (1) {
        if (datum[2]!='.' || datum[5]!='.') break;
        if (uhrzeit[2]!=':' || uhrzeit[5]!=':') break;

        t->tm_year = atoi(datum.substr(6, 4).c_str())-1900;
        t->tm_mon  = atoi(datum.substr(3, 2).c_str())-1;
        t->tm_mday = atoi(datum.substr(0, 2).c_str());

        t->tm_hour = atoi(uhrzeit.substr(0, 2).c_str());
        t->tm_min  = atoi(uhrzeit.substr(3, 2).c_str());
        t->tm_sec  = atoi(uhrzeit.substr(6, 2).c_str());

        t->tm_wday  = -1;
        t->tm_yday  = -1;
        t->tm_isdst = -1;

#ifdef MKTIME_TEST
        printf("-------------------- before mktime [2]:\n");
        printf("datum='%s'\nuhrzeit='%s'\n", datum.c_str(), uhrzeit.c_str());
        printf("t->tm_sec=%i\n", t->tm_sec);
        printf("t->tm_min=%i\n", t->tm_min);
        printf("t->tm_hour=%i\n", t->tm_hour);
        printf("t->tm_mday=%i\n", t->tm_mday);
        printf("t->tm_mon=%i\n", t->tm_mon);
        printf("t->tm_year=%i\n", t->tm_year);
        printf("t->tm_wday=%i\n", t->tm_wday);
        printf("t->tm_yday=%i\n", t->tm_yday);
        printf("t->tm_isdst=%i\n", t->tm_isdst);
#endif //MKTIME_TEST

        time_t ti = mktime(t);

#ifdef MKTIME_TEST
        printf("-------------------- after mktime [2]:\n");
        printf("ti=%lu\n", (long)ti);
        printf("t->tm_sec=%i\n", t->tm_sec);
        printf("t->tm_min=%i\n", t->tm_min);
        printf("t->tm_hour=%i\n", t->tm_hour);
        printf("t->tm_mday=%i\n", t->tm_mday);
        printf("t->tm_mon=%i\n", t->tm_mon);
        printf("t->tm_year=%i\n", t->tm_year);
        printf("t->tm_wday=%i\n", t->tm_wday);
        printf("t->tm_yday=%i\n", t->tm_yday);
        printf("t->tm_isdst=%i\n", t->tm_isdst);
#endif //MKTIME_TEST

        //         if (ti!=-1 && correct_date) {
        //             t->tm_mday++;
        //             ti = mktime(t);
        //         }
        if (ti==-1) break;

        return;
    }

    throw Error(string("Illegal date or time ('")+datum+"', '"+uhrzeit+"')");
}

// ----------------------------------------------------------
//     static string extrahiere_zugang(const string& s)
// ----------------------------------------------------------
static string extrahiere_zugang(const string& s) {
    size_t hk1 = s.find('\'');

    if (hk1!=string::npos) {
        size_t hk2 = s.find('\'', hk1+1);
        if (hk2!=string::npos) {
            return string(s, hk1+1, hk2-hk1-1);
        }
    }

    return "";
}

//  -----------------------------------
//      struct GetakteteVerbindung
//  -----------------------------------
struct GetakteteVerbindung {
    VerbindungPtr             v;
    const Verbindungsgebuehr *g;
    GetakteteVerbindung(VerbindungPtr v_, const Verbindungsgebuehr *g_) : v(v_), g(g_) {}
};

//  -----------------------------------------------------
//      static void calcCurrentCosts(InifilePtr ini)
//  -----------------------------------------------------
static void calcCurrentCosts(InifilePtr ini) {
    VerbindungenPtr verb = getVerbindungen(ini);
    if (verb->size()==0) throw("Es sind bisher keine Verbindungen definiert.");

    multimap<double, GetakteteVerbindung> byCost;
    int                                   seconds = atoi(forTime.get().c_str());

    for (Verbindungen::const_iterator vi=verb->begin(); vi!=verb->end(); ++vi) {

        VerbindungPtr v      = vi->second;
        struct tm     now    = global.today.get_Date();
        double        factor = v->getFaktor();

        if (!(factor>99.0)) {
            double cost = v->getCostFor(&now, seconds) * factor; // faktor (z.B. bei schnelleren Verbindungen wg. Kompression)
            byCost.insert(pair<double, GetakteteVerbindung>(cost, GetakteteVerbindung(v, Verbindung::getLastUsed())));
        }
    }

    for (multimap<double, GetakteteVerbindung>::const_iterator bc=byCost.begin(); bc!=byCost.end(); ++bc) {
        const GetakteteVerbindung& gv = bc->second;
        cout << "\"cost=" << bc->first << "\" \"takt=" << gv.g->getTakt() << "\" \"verbindung=" << gv.v->getName() << "\"\n";
    }
}

//  --------------------------------------------------------
//      static void onlineCheckInternal(InifilePtr ini)
//  --------------------------------------------------------
static void onlineCheckInternal(InifilePtr ini, string& momentaner_zugang, bool& online) {
    string     eventDatenbank = ini->findStringEntry("Allgemein", "FritzWebDatenbank");
    DBF_Reader db(eventDatenbank);
    int        last           = db.Records()-1;
    int        first          = last-20; // read last 20 records

    momentaner_zugang = "";
    online            = false;

#if defined(CHECK_FRIWEB_DBF)
    DBF_Reader db2("C:\\WINDOWS\\Anwendungsdaten\\Fritz!\\FRIWEB.DBF");
    db.dumpEntryDefinition(cout);
    cout << "----------------------------------------\n";
    db2.dumpEntryDefinition(cout);

#endif // CHECK_FRIWEB_DBF

    for (int current_record = first; current_record<=last; ++current_record) {
        db.gotoRecord(current_record);
        string info    = db.FieldContent("INFO");
        string meldung = db.FieldContent("MELDUNG");

#if defined(CHECK_FRIWEB_DBF)
        {
            string zeit  = db.FieldContent("ZEIT");
            string datum = db.FieldContent("DATUM");

            cout << "INFO=" << info << "\n";
            cout << "MELDUNG=" << meldung << "\n";
            cout << "ZEIT=" << zeit << "\n";
            cout << "DATUM=" << datum << "\n";
            cout << "----------------------------------------\n";
        }
#endif // CHECK_FRIWEB_DBF

        if (meldung.find("Zugang")!=string::npos && meldung.find("eingestellt")!=string::npos) {
            string zugang = extrahiere_zugang(meldung);
            if (zugang.empty()) {
                if (meldung.find("Kein")!=string::npos) {
                    momentaner_zugang = "";
                }
            }
            else {
                momentaner_zugang = zugang;
            }
            online = false;
        }
        else if (meldung.find("Verbindung zu")!=string::npos) {
            if (meldung.find("ist aufgebaut")!=string::npos) {
                string zugang = extrahiere_zugang(meldung);

                if (!zugang.empty()) {
                    online            = true;
                    momentaner_zugang = zugang;
                }
                else {
                    online            = false;
                    momentaner_zugang = "";
                }
            }
            else if (meldung.find("abgebaut")!=string::npos) {
                string zugang = extrahiere_zugang(meldung);

                online = false;
                if (!zugang.empty()) {
                    momentaner_zugang = zugang;
                }
                else {
                    momentaner_zugang = "";
                }
            }
        }

#if defined(DEBUG)
        cout << "momentaner_zugang=" << momentaner_zugang << "\n";
        cout << "online=" << int(online) << "\n";
        cout << "----------------------------------------\n";
#endif // DEBUG
    }

#if defined(CHECK_FRIWEB_DBF)
    int last2  = db2.Records()-1;
    int first2 = last2-20;      // read last 20 records

    for (int current_record = first2; current_record<=last2; ++current_record) {
        db2.gotoRecord(current_record);
        string BEZCHNG   = db2.FieldContent("BEZCHNG"); cout << "BEZCHNG=" << BEZCHNG << "\n";
        string CALLS     = db2.FieldContent("CALLS"); cout << "CALLS=" << CALLS << "\n";
        string DATUM     = db2.FieldContent("DATUM"); cout << "DATUM=" << DATUM << "\n";
        string DAUER     = db2.FieldContent("DAUER"); cout << "DAUER=" << DAUER << "\n";
        string ERROR     = db2.FieldContent("ERROR"); cout << "ERROR=" << ERROR << "\n";
        string GEBUEHREN = db2.FieldContent("GEBUEHREN"); cout << "GEBUEHREN=" << GEBUEHREN << "\n";
        string ID        = db2.FieldContent("ID"); cout << "ID=" << ID << "\n";
        string MELDUNG   = db2.FieldContent("MELDUNG"); cout << "MELDUNG=" << MELDUNG << "\n";
        string ONLINE    = db2.FieldContent("ONLINE"); cout << "ONLINE=" << ONLINE << "\n";
        string RUFNUMMER = db2.FieldContent("RUFNUMMER"); cout << "RUFNUMMER=" << RUFNUMMER << "\n";
        string ZEIT      = db2.FieldContent("ZEIT"); cout << "ZEIT=" << ZEIT << "\n";
        cout << "----------------------------------------\n";
    }
#endif // CHECK_FRIWEB_DBF
}
//  ------------------------------------------------
//      static void onlineCheck(InifilePtr ini)
//  ------------------------------------------------
static void onlineCheck(InifilePtr ini) {
    string zugang;
    bool   online;

    onlineCheckInternal(ini, zugang, online);

    //  ------------------------
    //      write online.ini
    //  ------------------------
    ofstream out(testOnline.get().c_str());

    out << "[All]\n";
    out << "zugang=" << zugang << "\n";
    out << "online=" << int(online) << "\n";
}
//  --------------------------------------------------------------------------
//      static bool correctedLongConnections(ConnectionList& connections)
//  --------------------------------------------------------------------------
static bool correctedLongConnections(ConnectionList& connections) {
    bool           corrected   = false;
    ConnectionList changed;
    int            minDuration = atoi(queryLongConnections.get().c_str()) * 60;

    for (ConnectionList::iterator ci=connections.begin(); ci!=connections.end(); ++ci) {
        int duration = ci->getDuration();

        if (duration>=minDuration) {
            cout << "Lange Verbindung: " << dateString(ci->getStart()) << " " <<
                timeString(ci->getStart()) << " " << timeString(duration) << "\n";
            bool done = false;
            while (!done) {
                string answer = ask_User("Verbindung kuerzen um hh:mm:ss ( - = nicht aendern )");
                if (answer!="-") {
                    if (answer[2]==':' && answer[5]==':') {
                        const char *a       = answer.c_str();
                        int duration_offset = (atoi(a)*60 + atoi(a+3))*60 + atoi(a+6);

                        if (duration_offset>0 && duration_offset<duration) {
                            ci->setDuration(duration-duration_offset);
                            done = corrected = true;
                        }
                    }
                    if (!done) cout << "Ungltige Eingabe\n";
                }
                else {
                    done = true;
                }
            }


        }
    }

    if (corrected) {
        cout << "Sorting connections...\n";
        connections.sort();
        cout << "done\n";
    }

    return corrected;
}
//  ------------------------------------------------------------------------------------
//      inline bool identicalVerbindung(const Verbindung& v1, const Verbindung& v2)
//  ------------------------------------------------------------------------------------
inline bool identicalVerbindung(const Verbindung& v1, const Verbindung& v2) {
    return v1.getName()==v2.getName();
}
//  ----------------------------------------------------------------------------
//      inline bool identicalTime(const struct tm *t1, const struct tm *t2)
//  ----------------------------------------------------------------------------
inline bool identicalTime(const struct tm *t1, const struct tm *t2) {
    return
        t1->tm_sec  == t2->tm_sec  &&
        t1->tm_min  == t2->tm_min  &&
        t1->tm_hour == t2->tm_hour &&
        t1->tm_mday == t2->tm_mday &&
        t1->tm_mon  == t2->tm_mon  &&
        t1->tm_year == t2->tm_year;
}
//  -------------------------------------------------------------------------------------
//      inline bool identicalConnections(const Connection& c1, const Connection& c2)
//  -------------------------------------------------------------------------------------
inline bool identicalConnections(const Connection& c1, const Connection& c2) {
    return
        c1.getKanal()    == c2.getKanal()                           &&
        c1.getDuration() == c2.getDuration()                        &&
        identicalVerbindung(c1.getVerbindung(), c2.getVerbindung()) &&
        identicalTime(c1.getStart(), c2.getStart());
}

static const char *wochentag[] = {
    "Sonntag   ",
    "Montag    ",
    "Dienstag  ",
    "Mittwoch  ",
    "Donnerstag",
    "Freitag   ",
    "Samstag   "
};

// ---------------------------------------------------------------
//      static size_t wochentagKategorie(const Connection& c)
// ---------------------------------------------------------------
static size_t wochentagKategorie(const Connection& c) {
    return c.getStart()->tm_wday;
}
// ----------------------------------------------------------------------------
//      static void writeWochentagKategorie(ostream& out, size_t category)
// ----------------------------------------------------------------------------
static void writeWochentagKategorie(ostream& out, size_t category) {
    out << wochentag[category];
}

// -------------------------------------------------------------
//      static size_t uhrzeitKategorie(const Connection& c)
// -------------------------------------------------------------
static size_t uhrzeitKategorie(const Connection& c) {
    return c.getStart()->tm_hour;
}
// --------------------------------------------------------------------------
//      static void writeUhrzeitKategorie(ostream& out, size_t category)
// --------------------------------------------------------------------------
static void writeUhrzeitKategorie(ostream& out, size_t category) {
    out << strf("%2i:oo Uhr", category);
}
//  ------------------------------------------------------------------------------
//      static void costStatistics(InifilePtr ini, const string& inifilename)
//  ------------------------------------------------------------------------------
static void costStatistics(InifilePtr ini, const string& inifilename) {
    VerbindungenPtr verb           = getVerbindungen(ini);
    string          eventDatenbank = ini->findStringEntry("Allgemein", "FritzWebDatenbank");
    ConnectionList  connections;

    {
        string oldConnectionsFileName = "fw_cost_old_connections.dat";
        bool   save_connections       = false;

        {
            bool connections_are_sorted;
            loadOldConnections(verb, connections, oldConnectionsFileName, connections_are_sorted);
            if (!connections_are_sorted) {
                connections.sort();
                save_connections = true;
            }
        }

        {
            ConnectionList    newConnections;
            bool              ini_changed = false;
            const Connection *last_old    = 0;

            if (connections.size()) {
                last_old = &connections.back();
            }

            readLog(newConnections, last_old, eventDatenbank, ini, verb, ini_changed);
            if (newConnections.size()) {
                newConnections.sort();
                save_connections = true;
                connections.merge(newConnections);
                connections.unique(identicalConnections);
            }

            if (ini_changed) writeInifile(inifilename, ini);
        }

        if (queryLongConnections) {
            save_connections = correctedLongConnections(connections);
        }

        if (save_connections) {
            saveOldConnections(verb, connections, oldConnectionsFileName);
        }
    }

    if (!queryLongConnections) {
        interpretConnections(connections, (bool)bundles ? atoi(maxPause.get().c_str()) : 0,
                             allowConnectionChange, einzel, detailed, provider);

        string von_bis = Zeitraum.get();
        if (von_bis!="all") {
            section_______________________(string("Gesamtstatistik fr Zeitraum ")+von_bis);

            struct tm von;
            struct tm bis;

            if (von_bis[10]!='-') throw Error(string("Illegal time range ('")+von_bis+"')");
            parseDateTime(&von, von_bis.substr(0, 10), "00:00:00");
            parseDateTime(&bis, von_bis.substr(11, 10), "23:59:59");

            time_t v = mktime(&von);
            time_t b = mktime(&bis);

            printConnectionSums(verb, connections, v, b, von_bis);
        }

        section_______________________("Zeitraum-Summen");
        double mean;
        if (tagessummen)  printSums(cout, connections, "Tagessummen", sameDay, writeDay, mean, 7, 30, provider);
        if (wochensummen) printSums(cout, connections, "Wochensummen", sameWeek, writeWeek, mean, 5, 13, provider);
        if (monatssummen) printSums(cout, connections, "Monatssummen", sameMonth, writeMonth, mean, 3, 12, provider);

        if (wochentagStatistik) printCategorizedSums(cout, connections, "Nach Wochentagen", wochentagKategorie, writeWochentagKategorie, provider);
        if (wochentagStatistik) printCategorizedSums(cout, connections, "Nach Uhrzeit", uhrzeitKategorie, writeUhrzeitKategorie, provider);
    }
}

//  ----------------------------------------------------------------------------
//      void deteriorateProvider(InifilePtr ini, const string& inifilename)
//  ----------------------------------------------------------------------------
void deteriorateProvider(InifilePtr ini, const string& inifilename) {
    string zugang;
    bool   online;

    onlineCheckInternal(ini, zugang, online);

    if (zugang == "") throw Error("No provider selected or FritzWeb not running.");

    SectionPtr   verb   = ini->findSection(zugang);
    EntryPtr     entry  = verb->findOrCreateEntry("Faktor", "1.0");
//     StringEntry& sentry = dynamic_cast<StringEntry&>(*entry);
    double       factor = strtod(entry->getContent().c_str(), 0);

    factor = factor + strtod(providerSucks.get().c_str(), 0);
    char buf[50];
    sprintf(buf, "%f", factor);
    entry->setContent(buf);

    writeInifile(inifilename, ini);
}

//  -----------------------------------------
//      int main(int argc, char *argv[])
//  -----------------------------------------
int main(int argc, char *argv[]) {
    try {
        fw_cost_info.init(argc, argv);
        if (!fw_cost_info.optionsSpecified()) fw_cost_info.throw_Error("Please specify an option.");

        {
            CmdLineOption* stat_options[] = { &einzel, &provider,
                                                  &tagessummen, &wochensummen, &monatssummen,
                                                  &bundles, &Zeitraum,
                                                  &wochentagStatistik, &stundenStatistik,
                                                  0 };
            CmdLineOption* online_options[]    = { &testOnline, 0 };
            CmdLineOption* currCost_options[]  = { &currentCosts, &forTime, 0 };
            CmdLineOption* queryLong_options[] = { &queryLongConnections, 0 };

            fw_cost_info.exclude(stat_options, online_options);
            fw_cost_info.exclude(stat_options, currCost_options);
            fw_cost_info.exclude(stat_options, queryLong_options);
            fw_cost_info.exclude(online_options, currCost_options);
            fw_cost_info.exclude(online_options, queryLong_options);
            fw_cost_info.exclude(currCost_options, queryLong_options);
        }

        // stat_options:
        fw_cost_info.neededBy(bundles, maxPause);
        fw_cost_info.neededBy(bundles, allowConnectionChange);
        fw_cost_info.neededBy(bundles, detailed);

        // currCost_options:
        fw_cost_info.neededBy(currentCosts, forTime);

        string inifilename = iniFile.get(); //"fw_cost.ini";
        if (!testOnline && !currentCosts) section_______________________(string("Lese ")+inifilename);
        InifilePtr ini     = readOrCreateInifile(inifilename, completeInifile);
        ini->fixState();

        if (testOnline) {
            onlineCheck(ini);
        }
        else if (currentCosts) {
            calcCurrentCosts(ini);
        }
        else if (providerSucks) {
            deteriorateProvider(ini, inifilename);
        }
        else {
            costStatistics(ini, inifilename);
        }

        return EXIT_SUCCESS;
    }
	catch (exception& e) {
        handle_exception(e);
	}
	catch (...) {
		Error err("Unknown exception");
		err.print(cerr);
	}

    return EXIT_FAILURE;
}



