/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sat Dec/07/2002 14:39 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

// #include <cstdlib>
#include <dbf.h>

#include "read.h"
#include "fw_cost.h"

using namespace std;
using namespace rs;
using namespace rs::ini;
using namespace rs::dbf;
using namespace rs::err;
using namespace rs::str;

//  ------------------------------------------------------------------
//      static string parseVerbindungsName(const string& meldung)
//  ------------------------------------------------------------------
static string parseVerbindungsName(const string& meldung) {
    string s("Verbindung zu '");
    size_t start = meldung.find(s);

    if (start!=string::npos) {
        start      += s.length();
        size_t end  = meldung.find('\'', start);
        if (end!=string::npos) {
            return meldung.substr(start, end-start);
        }
    }
    throw Error(string("No connection found in '")+meldung+'\'');
}

//  ---------------------------------------------------------------------------------------------------------------
//      static int cmpRecord(DBF_Reader& db, int record, const string& last_datum, const string& last_uhrzeit)
//  ---------------------------------------------------------------------------------------------------------------
inline int cmpRecord(DBF_Reader& db, int record, const string& last_datum, const string& last_uhrzeit) {
    // liefert  <0 falls last_...  < ...     (... steht fr datum/uhrzeit)
    // liefert ==0 falls last_... == ...
    // liefert  >0 falls last_...  > ...

//     cout << "Comparing record " << record;

    db.gotoRecord(record);
    string datum = db.FieldContent("DATUM");
    int    cmp;

//     cout << " datum=" << datum;

    if (last_datum==datum) {
        string uhrzeit = db.FieldContent("ZEIT");
//         cout << " uhrzeit=" << uhrzeit;
        cmp            = last_uhrzeit==uhrzeit ? 0 : (last_uhrzeit<uhrzeit ? -1 : 1);
    }
    else {
        cmp = last_datum<datum ? -1 : 1;
    }

//     cout << " cmp=" << cmp << "\n";

    return cmp;
}

namespace fw_cost {

    //  ---------------------------------------------------------------------------------------------------------------------------------------------------------------------
    //      void readLog(ConnectionList& connections, const Connection *lastLoaded, const string& databaseName, InifilePtr ini, VerbindungenPtr verb, bool& ini_changed)
    //  ---------------------------------------------------------------------------------------------------------------------------------------------------------------------
    void readLog(ConnectionList& connections, const Connection *lastLoaded, const string& databaseName, InifilePtr ini, VerbindungenPtr verb, bool& ini_changed) {
        try {
            section_______________________(string("Lese ")+databaseName);

            DBF_Reader db(databaseName);
            Kanal      kanal[2];
            bool       last_was_open_cmd = false;
            string     last_datum;
            string     last_uhrzeit;
            bool       done              = false;

            if (lastLoaded) {   // we have loaded old connections
                struct tm t_conn;
                lastLoaded->getEnd(&t_conn);

                string lastConn_datum   = rs::str::strf("%04i%02i%02i", t_conn.tm_year+1900, t_conn.tm_mon+1, t_conn.tm_mday);
                string lastConn_uhrzeit = rs::str::strf("%02i%02i%02i", t_conn.tm_hour, t_conn.tm_min, t_conn.tm_sec);

//                 cout << "searching entry for lastConn_datum=" << lastConn_datum << " lastConn_uhrzeit=" << lastConn_uhrzeit << "\n";

                int records = db.Records();
                int high    = records-1;
                int low     = 0;
                int found   = -1;
                int cmp;

                cmp = cmpRecord(db, low, lastConn_datum, lastConn_uhrzeit);
                if (cmp==0) found = low;
                else {
                    cmp = cmpRecord(db, high, lastConn_datum, lastConn_uhrzeit);
                    if (cmp==0) found = high;
                    else {
                        while (found==-1 && low<(high-1)) {
                            int middle = (low+high)/2;
                            cmp        = cmpRecord(db, middle, lastConn_datum, lastConn_uhrzeit);

                            if (cmp==0) found = middle;
                            else {
                                if (cmp<0) high = middle;
                                else       low = middle;
                            }
                        }
                    }
                }

                if (found!=-1) {
                    cout << "Suche nach neuen Eintrgen in der Fritzweb Datenbank...\n";

                    cmp = 0;
                    while (cmp==0) {
                        ++found;
                        if (found>=records) {
                            done = true; // no new records
                            break;
                        }
                        cmp = cmpRecord(db, found, lastConn_datum, lastConn_uhrzeit);
                    }
                }
                else {
                    found = 0;
                    db.gotoRecord(0);
                }

                if (!done) {
                    cout << "- Lese Eintrge " << found << " bis " << records << "\n";
                }
                else {
                    cout << "- Datenbank bereits vollstaendig gelesen.\n";
                }
            }

            while (!done) {
                string meldung = db.FieldContent("MELDUNG");
#if defined(DUMP_DBF_PARSING)
                cout << "meldung='" << meldung << "'\n";
#endif // DUMP_DBF_PARSING
                bool this_was_open_cmd = false;

                string isdn("N-Kanal"); // AVM produziert Schreibfehler (IDSN und ISDN)
                size_t kanal_meldung = meldung.find(isdn);

                if (kanal_meldung==string::npos) {
                    bool   close_channels = false;
                    string reason;

                    if (meldung=="System wird initialisiert.") {
                        if (kanal[0].is_open() || kanal[1].is_open()) {
                            reason         = "Kanal offen bei FritzWeb-Initialisierung";
                            close_channels = true;
                        }
                    }
                    else if (meldung=="System bereit. Kein Zugang eingestellt.") {
                        if (kanal[0].is_open() || kanal[1].is_open()) {
                            reason         = "Fritzweb deaktiviert";
                            close_channels = true;
                        }
                    }

                    if (close_channels) {
                        string datum   = db.FieldContent("DATUM");
                        string uhrzeit = db.FieldContent("ZEIT");

                        cout << reason << " (Datum=" << datum << " Uhrzeit=" << uhrzeit << ")\n";

                        for (int kanal_nummer=0; kanal_nummer<=1; ++kanal_nummer) {
                            if (kanal[kanal_nummer].is_open()) {
                                const string& verb_name = kanal[kanal_nummer].VerbundenMit();
                                VerbindungPtr mit       = findVerbindung(verb, verb_name, ini_changed);

                                kanal[kanal_nummer].close(connections, mit, datum, uhrzeit, kanal_nummer);
                            }
                        }

                    }
                }
                else {          // meldung enthaelt N-Kanal (ISDN-Kanal)
                    int    kanal_nummer  = meldung[kanal_meldung+isdn.length()+1]-'1';
                    string verbunden_mit = parseVerbindungsName(meldung);

                    if (verbunden_mit.length()==0) {
                        string datum   = db.FieldContent("DATUM");
                        string uhrzeit = db.FieldContent("ZEIT");

                        cout << "Kein Verbindungsname (Reset des Rechners whrend aktiver Verbindung?)."
                             << " (Datum=" << datum << ", Uhrzeit=" << uhrzeit << ")\n";
                        if (kanal[kanal_nummer].is_open()) {
                            const string& verb_name = kanal[kanal_nummer].VerbundenMit();
                            VerbindungPtr mit       = findVerbindung(verb, verb_name, ini_changed);

                            kanal[kanal_nummer].close(connections, mit, datum, uhrzeit, kanal_nummer);
                        }
                    }
                    else {
                        if (kanal_nummer<0 || kanal_nummer>1)
                            throw Error(string("Unexpected ISDN channel ")+char('1'+kanal_nummer));

                        string datum   = db.FieldContent("DATUM");
                        string uhrzeit = db.FieldContent("ZEIT");
#if defined(DUMP_DBF_PARSING)
                        cout << "Datum='" << datum << "' Uhrzeit='" << uhrzeit << "'\n";
#endif // DUMP_DBF_PARSING

#ifdef DATE_CORRECTION
                        if (datum==last_datum && uhrzeit<last_uhrzeit) { // datums-umschaltung fehlgeschlagen
                            Kanal::set_date_corrected(true);
                            cout << "Datums-Korrektur aktiviert!\n";
                        }
                        else if (Kanal::date_is_corrected() && datum>last_datum && uhrzeit>last_uhrzeit) {
                            // datum stimmt jetzt wieder
                            Kanal::set_date_corrected(false);
                            cout << "Datums-Korrektur deaktiviert!\n";
                        }
#endif // DATE_CORRECTION

                        last_datum   = datum;
                        last_uhrzeit = uhrzeit;

                        if (meldung.find("aufgebaut")!=string::npos) {
                            this_was_open_cmd = true;
                            if (!(meldung.find("ist aufgebaut") &&
                                  kanal[kanal_nummer].is_open() &&
                                  last_was_open_cmd)) { // nicht bereits durch 'wird aufgebaut' geoeffnet

                                if (kanal[kanal_nummer].is_open()) {
                                    cout << "Kanal war bereits offen - schliesse alle Kanaele."
                                         << " (Datum=" << datum << ", Uhrzeit=" << uhrzeit << ")\n";
                                    VerbindungPtr mit;

                                    if (kanal_nummer==0 && kanal[1].is_open()) { // zweiten Kanal auch schliessen
                                        mit = findVerbindung(verb, kanal[1].VerbundenMit(), ini_changed);
                                        kanal[1].close(connections, mit, datum, uhrzeit, 1);
                                    }

                                    mit = findVerbindung(verb, kanal[kanal_nummer].VerbundenMit(), ini_changed);
                                    kanal[kanal_nummer].close(connections, mit, datum, uhrzeit, kanal_nummer);
                                }
                                assert(kanal[kanal_nummer].is_closed());
                                kanal[kanal_nummer].open(verbunden_mit, datum, uhrzeit);
                            }
                        }
                        else if (meldung.find("abgebaut")!=string::npos) {
                            VerbindungPtr mit = findVerbindung(verb, verbunden_mit, ini_changed);

                            if (kanal[kanal_nummer].is_open()) {
                                kanal[kanal_nummer].close(connections, mit, datum, uhrzeit, kanal_nummer);
                            }
                            else {
                                cout << "Kanal war gar nicht offen\n";
                            }
                        }
                    }
                }

                last_was_open_cmd = this_was_open_cmd;
                done              = db.nextRecord()==0;
            }
        }
        catch (Error& err) {
            throw Error(string("While reading database '")+databaseName+"': "+err.get_message());
        }

#if defined(DEBUG)
//         cout << *ini;
#endif // DEBUG

        if (ini_changed) {
            SectionPtr sec_verbindungen = ini->findSection("Verbindungen");
            sec_verbindungen->clearEntries(); // alte Verbindungen loeschen
            int        num              = 0;

            // und dann alle Verbindungen eintragen
            for (Verbindungen::iterator v=verb->begin(); v != verb->end(); ++v) {

                VerbindungPtr verbindung = v->second;
                string        entryName  = getConnectionEntryName(num++);

                sec_verbindungen->insertEntry(new Entry(entryName, verbindung->getName()));

                SectionPtr sec_verbindung = ini->findOrCreateSection(verbindung->getName());
#if defined(DEBUG)
//                 cout << *sec_verbindung;
#endif // DEBUG
                sec_verbindung->clearEntries();
                sec_verbindung->findOrCreateEntry("Faktor", rs::str::strf("%f", verbindung->getFaktor()));

                // Einzelne Taktungen rueckwaerts speichern
                unsigned num2 = 0;
                for (list<Verbindungsgebuehr>::const_reverse_iterator g=verbindung->rbegin();
                     g!=verbindung->rend();
                     ++g, ++num2)
                {
                    string entryName2 = getTaktungEntryName(num2);
                    sec_verbindung->findOrCreateEntry(entryName2+"_Zeitraum", g->getZeitraum());
                    sec_verbindung->findOrCreateEntry(entryName2+"_Taktung", g->getTaktung());
                }
            }
        }
    }

};

