/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sat May/04/2002 11:42 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#include <BinFile.h>

#include "fw_cost.h"
#include "store.h"

using namespace std;
using namespace rs;
using namespace rs::err;
using namespace rs::file;

inline void put_struct_tm(const struct tm& tm1, BinaryFile& file) {
    struct tm tm2 = tm1;
    time_t    t   = mktime(&tm2);
    assert(t!=-1);
    put_binary(t, file);
}
inline void get_struct_tm(struct tm& tm1, BinaryFile& file) {
    time_t t;
    get_binary(t, file);
    tm1 = *localtime(&t);
}

namespace fw_cost {

    const string KENNUNG = "FW_COST_SAVE";
    const int VERSION    = 2;

    //  -----------------------------------------------------------------------------------------------------------
    //      void saveOldConnections(VerbindungenPtr verb, ConnectionList& connections, const string& filename)
    //  -----------------------------------------------------------------------------------------------------------
    void saveOldConnections(VerbindungenPtr verb, ConnectionList& connections, const string& filename) {
        section_______________________(string("Schreibe ")+filename);
        //ofstream out(filename.c_str(), ios::binary|ios::out);
        BinaryFile out(filename, BF_OUTPUT);
        if (!out.good()) throw IOError(filename);

        map<string, int> verb_nummer; // string == long name

        out.put_string(KENNUNG);
        out.put_int(VERSION);

        // Verbindungen speichern (nur Name)
        out.put_int(verb->size());
        int vcount = 0;
        for (Verbindungen::const_iterator v=verb->begin(); v!=verb->end(); ++v) {
            verb_nummer[v->first] = vcount++;
            out.put_string(v->first);
        }

        // Connections speichern
        out.put_int(connections.size());
        for (ConnectionList::const_iterator c=connections.begin(); c!=connections.end(); ++c) {
            const Verbindung& v = c->getVerbindung();
            out.put_int(verb_nummer[v.getName()]);
            put_struct_tm(*(c->getStart()), out);
            out.put_int(c->getDuration());
            out.put_int(c->getKanal());
        }
    }
    //  ----------------------------------------------------------------------------------------------------------------------------
    //      void loadOldConnections(VerbindungenPtr verb, ConnectionList& connections, const string& filename, bool& is_sorted)
    //  ----------------------------------------------------------------------------------------------------------------------------
    void loadOldConnections(VerbindungenPtr verb, ConnectionList& connections, const string& filename, bool& is_sorted) {
        section_______________________(string("Lese ")+filename);
        BinaryFile *in = 0;
        try {
            in = new BinaryFile(filename, BF_INPUT);
        }
        catch (...) {
            is_sorted = true;   // file does not exist -> empty array is always sorted
            return;
        }
        assert(in);

        if (in->good()) {
            {
                string kennung;
                in->get_string(kennung);
                //                 cout << "kennung='" << kennung << "'\n";
                //                 cout << "KENNUNG='" << KENNUNG << "'\n";
                if (kennung!=KENNUNG) throw Error(string("'")+filename+"' has wrong format");
            }

            int version; in->get_int(version);
            is_sorted = (version>=2); // before version 2 connections were not sorted

            // Verbindungen lesen
            int verbindungen; in->get_int(verbindungen);
            string *verb_name = new string[verbindungen];
            for (int v=0; v<verbindungen; ++v) in->get_string(verb_name[v]);

            // Connections lesen
            int conn_count; in->get_int(conn_count);

            for (int c=0; c<conn_count; ++c) {
                int       vnum; in->get_int(vnum);
                struct tm t; get_struct_tm(t, *in);
                int       duration; in->get_int(duration);
                int       kanal; in->get_int(kanal);

                assert(vnum>=0 && vnum<verbindungen);
                string        name = verb_name[vnum];
                VerbindungPtr v    = (*verb)[name];
                assert(!v.Null());

                connections.push_back(Connection(v, t, duration, kanal));
            }
        }
        delete in;
    }

}; // end of namespace fw_cost


